from pyxai.source.core.tools.encoding import CNFencoding
from pyxai.source.core.structure.type import OperatorCondition

class BinaryMapping():

  def __init__(self, map_id_binaries_to_features, map_features_to_id_binaries, ML_solver_information):
    self.map_id_binaries_to_features = map_id_binaries_to_features
    self.map_features_to_id_binaries = map_features_to_id_binaries
    self.ML_solver_information = ML_solver_information

  def compute_id_binaries(self):
    assert False, "Have to be implemented in a child class." 
  
  def instance_to_binaries(self, instance, preference_order=None):
    """
    map_id_binaries_to_features: list[id_binary] -> (id_feature, operator, threshold) 
    map_features_to_id_binaries: dict[(id_feature, operator, threshold)] -> [id_binary, n_appears, n_appears_per_tree]
    Transform a instance into a cube (conjunction of literals) according to the tree
    """
    output = []
    if preference_order is None:
      for key in self.map_features_to_id_binaries.keys(): # the keys are of the type: (id_feature, operator, threshold)
        id_feature = key[0]
        operator = key[1]
        threshold = key[2]

        if operator == OperatorCondition.GE:
          sign = 1 if instance[id_feature-1] >= threshold else -1
        elif operator == OperatorCondition.GT:
          sign = 1 if instance[id_feature-1] > threshold else -1
        elif operator == OperatorCondition.LE:
          sign = 1 if instance[id_feature-1] <= threshold else -1
        elif operator == OperatorCondition.LT:
          sign = 1 if instance[id_feature-1] < threshold else -1
        elif operator == OperatorCondition.EQ:
          sign = 1 if instance[id_feature-1] == threshold else -1
        elif operator == OperatorCondition.NEQ:   
          sign = 1 if instance[id_feature-1] != threshold else -1
        else:
          raise NotImplementedError("The operator "+str(self.operator)+ " is not implemented.")

        output.append(sign * self.map_features_to_id_binaries[key][0]) #map_features_to_id_binaries[key][0] => id_binary
      return CNFencoding.format(output)
    assert True, "To implement"
    return output
  

  def get_id_features(self, implicant):
    return tuple(self.map_id_binaries_to_features[abs(l)][0] for l in implicant)

  def to_features(self, implicant, eliminate_redundant_features=True, details=False):
    """
    Convert an implicant into features. Return a tuple of features. 
    Two types of features are available according to the details parameter. 
    If details is False, each feature is represented by a string "name >= threshold" or "name < threshold". 
    Else, if details is True, each feature is a python dict with the keys "id", "name", "threshold", "sign" and "weight".
    "id": The id of the feature (from 1 to the number of features) of the condition node represented by the literal of the implicant. 
    "name": The name of the feature. It is None if it is not available.
    "threshold": The threshold of the condition node represented by the literal of the implicant. 
    "sign": If the associated literal was positive in the implicant (resp. negative), it is True (resp. False). 
    "weight": To give a weight to the feature (optional)
    
    Remark: call self.eliminate_redundant_features()
    """
    result = []
    if eliminate_redundant_features:
      implicant = self.eliminate_redundant_features(implicant)
    for l in implicant:
      feature = dict()
      feature["id"] = self.map_id_binaries_to_features[abs(l)][0]

      if self.ML_solver_information is None:
        feature["name"] = "f"+str(feature["id"])
      elif self.ML_solver_information.feature_name is None:
        feature["name"] = "f"+str(feature["id"])
      else:
        feature["name"] = self.ML_solver_information.feature_name[feature["id"]-1]
      
      feature["operator"] = self.map_id_binaries_to_features[abs(l)][1]
      feature["threshold"] = self.map_id_binaries_to_features[abs(l)][2]
      feature["sign"] = True if l > 0 else False
      feature["weight"] = implicant[l] if isinstance(implicant, dict) else None
      if details:
        result.append(feature)
      else: 
        if feature["operator"] == OperatorCondition.GE:
          str_sign = " >= " if feature["sign"] else " < "
        elif feature["operator"] == OperatorCondition.GT:
          str_sign = " > " if feature["sign"] else " <= "
        elif feature["operator"] == OperatorCondition.LE:
          str_sign = " <= " if feature["sign"] else " > "
        elif feature["operator"] == OperatorCondition.LT:
          str_sign = " < " if feature["sign"] else " >= "
        elif feature["operator"] == OperatorCondition.EQ:
          str_sign = " == " if feature["sign"] else " != "
        elif feature["operator"] == OperatorCondition.NEQ:
          str_sign = " != " if feature["sign"] else " == "
        else:
          raise NotImplementedError("The operator "+str(self.operator)+ " is not implemented.")
        
        result.append(str(feature["name"])+str_sign+str(feature["threshold"]))
    return tuple(result)

  def extract_excluded_features(self, implicant, excluded_features):
    """Return index of the implicant to exclude. 

    Args:
        implicant (_type_): _description_
        excluded_features (_type_): _description_

    Returns:
        :obj:`list` of int: List of index of the implicant to exclude
    """
    return [i for i, feature in enumerate(self.to_features(implicant, eliminate_redundant_features=False, details=True)) if feature["id_feature"] in excluded_features]

  def eliminate_redundant_features(self, implicant):
    """
    A implicant without redundant features i.e. If we have 'feature_a > 3' and 'feature_a > 2', we keep only the id_binary linked to the boolean corresponding to the 'feature_a > 3'
    Warning, the 'implicant' parameter can be a list of literals or a dict of literals. 
    In the last case, it is a map literal -> weight.
    """
    positive = {}
    positive_weights = {}
    negative = {}
    negative_weights = {}
    n_redundant_features = 0
    no_done = []
    #Search redundant features
    for l in implicant:
      key = self.map_id_binaries_to_features[abs(l)]
      id_feature = key[0]
      operator = key[1]
      threshold = key[2]
      if operator == OperatorCondition.GE:
        if l > 0:
          if (id_feature, operator) not in positive:
            positive[(id_feature, operator)] = [threshold]
            if isinstance(implicant, dict):
              positive_weights[(id_feature, operator)] = implicant[l]
          else:
            n_redundant_features+=1
            positive[(id_feature, operator)].append(threshold)
            if isinstance(implicant, dict):
              positive_weights[(id_feature, operator)] += implicant[l]
        else:
          if (id_feature, operator) not in negative:
            negative[(id_feature, operator)] = [threshold]
            if isinstance(implicant, dict):
              negative_weights[(id_feature, operator)] = implicant[l]
          else:
            n_redundant_features+=1
            negative[(id_feature, operator)].append(threshold)
            if isinstance(implicant, dict):
              negative_weights[(id_feature, operator)] += implicant[l]
      else:
        no_done.append(l)
    #Copy the new implicant without these redundant features
    if not isinstance(implicant, dict):
      output = [self.map_features_to_id_binaries[(id, operator, max(positive[(id, operator)]))][0] for (id, operator) in positive.keys()]
      output += [-self.map_features_to_id_binaries[(id, operator, min(negative[(id, operator)]))][0] for (id, operator) in negative.keys()]
      output += no_done
      return tuple(output)
    
    output = [(self.map_features_to_id_binaries[(id, operator, max(positive[(id, operator)]))][0], positive_weights[(id, operator)]) for (id, operator) in positive.keys()]
    output += [(-self.map_features_to_id_binaries[(id, operator, min(negative[(id, operator)]))][0], negative_weights[(id, operator)]) for (id, operator) in negative.keys()]
    output += no_done
    output = {t[0]:t[1] for t in output} #To dict
    return output
    